
// Include libraries
#include <SoftwareSerial.h> // To communicate with the GPS
#include <Serial.h>
#include <Wire.h> // I2C communications (supporting the LSM303 library)
#include <LSM303.h> // Custom library for the compass board
#include <Flash.h> // Custom library for the NAND Flash
#include <SPI.h> // SPI communications (supporting the NAND-Flash library)
#include <EEPROM.h> // For non-volatile memory on the Arduino (supporting the NAND-Flash library)

#define SAMPLE_INTERVAL 10000 // 100 Hz
#define MAX_GPS_DATA 100 // 100 bytes
#define DEBUG // output extra status information via serial

// Analog pins
#define PIP_CTRL A0
#define PIP_IN A1

// Specify DNT pins
#define DNT_SLEEP 7 // Active High
#define DNT_RST 6 // Active Low
#define DNT_RTS 5 // Active High
#define DNT_CTS 4 // Set as input
#define DNT_RX 1
#define DNT_TX 0

Flash* nandFlash; // Pointer to nandflash object
SoftwareSerial GPS(3, 2); // GPS serial interface
LSM303 compass; // Compass object

// At the start of each line of data
byte GPS_SENTINEL = '$'; // To indicate GPS data
byte DATA_SENTINEL = '#'; // To indicate a line of compass data

void setup() {
  // Configure DNT pins
  pinMode(DNT_TX, OUTPUT);
  pinMode(DNT_RST, OUTPUT);
  digitalWrite(DNT_RST, LOW);
  digitalWrite(DNT_TX, LOW);
  delay(200);
  digitalWrite(DNT_RST, HIGH);
  delay(200);
  pinMode(DNT_TX, INPUT);
  
  pinMode(DNT_SLEEP, OUTPUT);
//  pinMode(DNT_RST, OUTPUT);
  pinMode(DNT_RTS, OUTPUT);
  pinMode(DNT_CTS, INPUT);
  pinMode(DNT_RX, OUTPUT);
//  pinMode(DNT_TX, INPUT);
  
  digitalWrite(DNT_SLEEP, LOW);
//  digitalWrite(DNT_RST, HIGH);
  digitalWrite(DNT_RTS, LOW);
  digitalWrite(DNT_CTS, LOW); // disable pullups

  delay(1000);
  Serial.begin(115200);
  nandFlash = new Flash(true, false); // Flash(enable_write, restart_address_counter)
  
  // Compass setup
  Wire.begin();
  compass.init(); //Turns on LSM303DLH
  compass.writeAccReg(LSM303_CTRL_REG1_A, 0x57); // Acc output data rate (ODR) = 100 Hz, not low power 
  compass.writeAccReg(LSM303_CTRL_REG4_A, 0x18); // Set accelerometer scale to +/-4G w HR
  compass.writeMagReg(LSM303_MR_REG_M, 0x00);    // Default
  compass.writeMagReg(LSM303_CRA_REG_M, 0x18);   // Mag output data rate (ODR) = 75 Hz
  compass.writeMagReg(LSM303_CRB_REG_M, 0x20);   // Mag gain +/-1.3 Gauss (X,Y - 1100 LSB/Gauss, Z - 980 LSB/Gauss)

  
  // Configue GPS
  delay(500);
  GPSsetup();
  delay(500);
  
  // adjust the prescale for the analog inputs (http://arduino.cc/forum/index.php/topic,6549.0.html)
  // defines for setting and clearing register bit
  #ifndef cbi
  #define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
  #endif
  #ifndef sbi
  #define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
  #endif
  
  // set prescale to 32
  sbi(ADCSRA,ADPS2) ;
  cbi(ADCSRA,ADPS1) ;
  sbi(ADCSRA,ADPS0) ; 
}

void loop() {
  // GPS data
  byte GPSdata[MAX_GPS_DATA]; // Temporary storage (max 100)
  byte GPSchar = 0; // New char
  int GPSindex = 0; // Current position in the GPS data array
  
  // Analong data (2) and compass data (6)
  int data[8];

  // Data time stamp (in milliseconds)
  unsigned long time = 0; // Time of last/current sample
//  unsigned long prev_time = 0; // Time of last mag sample
  
  while(1)
  { 
    // If the GPS has data ready...
    if(GPS.available())
    {
      // Get the next character
      GPSchar = GPS.read();
      
      // If in the middle of a line with memory still available, append data
      if ((GPSchar != GPS_SENTINEL) && (GPSindex < MAX_GPS_DATA))
      {
        GPSdata[GPSindex] = GPSchar;
        GPSindex+=1;
      }
      else // Either finished a line (starting a new one) or ran out of temporary storage space
      {
#ifdef DEBUG
        if (GPSchar != GPS_SENTINEL)
          Serial.println("Temporary storage space exceeded!");
#endif
        
        // Send complete string (or at least as much as you have)
        Serial.write(GPSdata,GPSindex);
        nandFlash->writeBytes(GPSdata,GPSindex);
        delayMicroseconds(10);
        
        GPSdata[0] = GPS_SENTINEL;

        long GPStime = micros(); // timestamp for the next string
        GPSdata[1] = (char)(GPStime & 0xFF);
        GPSdata[2] = (char)((GPStime >> 8) & 0xFF);
        GPSdata[3] = (char)((GPStime >> 16) & 0xFF);
        GPSdata[4] = (char)((GPStime >> 24) & 0xFF);
        
        GPSindex = 5; // Reset index counter
        
        time = micros(); // So the DNT doesn't get overrun
      }
    }
    
    // Control the sampling cadence
    if ((micros() - time) > SAMPLE_INTERVAL)
    {
      // Collect data as close to time stamp as possible
      time = micros();
      compass.read();
      data[0] = (unsigned int)analogRead(PIP_CTRL); //dummy read
      data[0] = (unsigned int)analogRead(PIP_CTRL); //dummy read
      data[0] = (unsigned int)analogRead(PIP_CTRL);
      data[2] = (int)compass.a.x;
      data[3] = (int)compass.a.y;
      data[4] = (int)compass.a.z;
      data[5] = (int)compass.m.x;
      data[6] = (int)compass.m.y;
      data[7] = (int)compass.m.z;
      data[1] = (unsigned int)analogRead(PIP_IN); //dummy read
      data[1] = (unsigned int)analogRead(PIP_IN); //dummy read
      data[1] = (unsigned int)analogRead(PIP_IN);
  
      // Send data over the radio
      Serial.write(DATA_SENTINEL);
      Serial.write((byte *)(&time),4);
      Serial.write((byte *)data,16);
      
      // Send it to the nandflash
      nandFlash->writeBytes(&DATA_SENTINEL,1);
      nandFlash->writeBytes((byte *)(&time),4);
      nandFlash->writeBytes((byte *)data,16);
                  
//      prev_time = time;
    }
  }
}

// ------------------------------ GPS Setup ------------------------------ //

void GPSsetup() {
  GPS.begin(9600);
  delay(500);
  
//  // Message to set the serial baud rate
//  uint8_t setBaud[] = {0x05, 0x00, 0x02, 0x00}; // COM 1, 19200 baud
//  
//  sendMSG(setBaud, sizeof(setBaud)/sizeof(uint8_t));
//  delay(500);
//  GPS.begin(19200);
//  GPS.flush();
  
  // Message to configure NMEA messages
  uint8_t setNMEAmsg[] = {0x08, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00}; // only GGA, once per second
  
  do {
#ifdef DEBUG
    Serial.println("GPS setup loop");
#endif
    sendMSG(setNMEAmsg, sizeof(setNMEAmsg)/sizeof(uint8_t));
  } while(!getACK(setNMEAmsg));
}

void sendMSG(uint8_t *msg, uint8_t len) {
  
  // Indicates the start of a message
  GPS.write(0xA0);
  GPS.write(0xA1);
  
  // Message length (big endian)
  GPS.write((uint8_t)0x0);
  GPS.write(len);
  
  // Send the message body
  uint8_t checksum = 0;
  for(int i=0; i<len; i++) {
    GPS.write(msg[i]);
    checksum = checksum ^ msg[i];
  }
  
  // Message body checksum
  GPS.write(checksum);
  
  // Indicates the end of the message
  GPS.write(0x0D);
  GPS.write(0x0A);
}
 
boolean getACK(uint8_t *msg) {
  uint8_t b;
  uint8_t ackByteID = 0;
  uint8_t ackPacket[9];
  unsigned long startTime = millis();
 
  // Construct the expected ACK packet    
  ackPacket[0] = 0xA0;	// header
  ackPacket[1] = 0xA1;	// header
  ackPacket[2] = 0x00;	// length
  ackPacket[3] = 0x02;	// length
  
  ackPacket[4] = 0x83;	// message ID
  ackPacket[5] = 0x00;//msg[0];// ID of the message it is acknowledging
  
  ackPacket[6] = 0x83;// ^ msg[0]; // checksum (xor of bytes in the message body)
  ackPacket[7] = 0x0D;	// footer
  ackPacket[8] = 0x0A;	// footer
 
  while (1) {
    // Test for success
    if (ackByteID > 8) {
      // All packets in order!
      return true;
    }
 
    // Timeout if no valid response in 3 seconds
    if (millis() - startTime > 3000) { 
      return false;
    }
 
    // Make sure data is available to read
    if (GPS.available()) {
      b = GPS.read();
      Serial.write(b);
 
      // Check that bytes arrive in sequence as per expected ACK packet
      if (b == ackPacket[ackByteID]) { 
        ackByteID++;
      } 
      else {
        ackByteID = 0;	// Reset and look again, invalid order
      }
    }
  }
}
